# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load, validate

from polyaxon_schemas.base import BaseConfig
from polyaxon_schemas.experiment import ExperimentSchema
from polyaxon_schemas.utils import UUID


class ExperimentGroupSchema(Schema):
    id = fields.Int(allow_none=True)
    uuid = UUID(allow_none=True)
    name = fields.Str(validate=validate.Regexp(regex=r'^[-a-zA-Z0-9_]+\Z'), allow_none=True)
    unique_name = fields.Str(allow_none=True)
    user = fields.Str(validate=validate.Regexp(regex=r'^[-a-zA-Z0-9_]+\Z'), allow_none=True)
    project = fields.Str(allow_none=True)
    description = fields.Str(allow_none=True)
    content = fields.Str()
    tags = fields.List(fields.Str(), allow_none=True)
    created_at = fields.LocalDateTime(allow_none=True)
    updated_at = fields.LocalDateTime(allow_none=True)
    concurrency = fields.Int(allow_none=True)
    num_experiments = fields.Int(allow_none=True)
    num_scheduled_experiments = fields.Int(allow_none=True)
    num_pending_experiments = fields.Int(allow_none=True)
    num_running_experiments = fields.Int(allow_none=True)
    num_succeeded_experiments = fields.Int(allow_none=True)
    num_failed_experiments = fields.Int(allow_none=True)
    num_stopped_experiments = fields.Int(allow_none=True)
    last_status = fields.Str(allow_none=True)
    has_tensorboard = fields.Bool(allow_none=True)
    experiments = fields.Nested(ExperimentSchema, many=True, allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ExperimentGroupConfig(**data)

    @post_dump
    def unmake(self, data):
        return ExperimentGroupConfig.remove_reduced_attrs(data)


class ExperimentGroupConfig(BaseConfig):
    SCHEMA = ExperimentGroupSchema
    IDENTIFIER = 'experiment_group'
    DEFAULT_INCLUDE_ATTRIBUTES = [
        'id', 'unique_name', 'user', 'concurrency', 'num_experiments',
        'num_pending_experiments', 'num_running_experiments', 'created_at', 'last_status'
    ]
    DATETIME_ATTRIBUTES = ['created_at', 'updated_at']

    def __init__(self,
                 unique_name=None,
                 id=None,  # pylint:disable=redefined-builtin
                 user=None,
                 name=None,
                 description=None,
                 content=None,
                 uuid=None,
                 project=None,
                 num_experiments=None,
                 tags=None,
                 num_scheduled_experiments=None,
                 num_pending_experiments=None,
                 num_running_experiments=None,
                 num_succeeded_experiments=None,
                 num_failed_experiments=None,
                 num_stopped_experiments=None,
                 last_status=None,
                 has_tensorboard=False,
                 created_at=None,
                 updated_at=None,
                 concurrency=None,
                 experiments=None):
        self.unique_name = unique_name
        self.id = id
        self.user = user
        self.name = name
        self.description = description
        self.content = content
        self.uuid = uuid
        self.project = project
        self.tags = tags
        self.num_experiments = num_experiments
        self.num_scheduled_experiments = num_scheduled_experiments
        self.num_pending_experiments = num_pending_experiments
        self.num_running_experiments = num_running_experiments
        self.num_succeeded_experiments = num_succeeded_experiments
        self.num_failed_experiments = num_failed_experiments
        self.num_stopped_experiments = num_stopped_experiments
        self.created_at = self.localize_date(created_at)
        self.updated_at = self.localize_date(updated_at)
        self.has_tensorboard = has_tensorboard
        self.last_status = last_status
        self.concurrency = concurrency
        self.experiments = experiments


class GroupStatusSchema(Schema):
    id = fields.Int()
    uuid = UUID()
    experiment_group = fields.Int()
    created_at = fields.LocalDateTime()
    status = fields.Str()
    message = fields.Str(allow_none=True)
    details = fields.Dict(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return GroupStatusConfig(**data)

    @post_dump
    def unmake(self, data):
        return GroupStatusConfig.remove_reduced_attrs(data)


class GroupStatusConfig(BaseConfig):
    SCHEMA = GroupStatusSchema
    IDENTIFIER = 'GroupStatus'
    DATETIME_ATTRIBUTES = ['created_at']
    DEFAULT_EXCLUDE_ATTRIBUTES = ['experiment_group', 'uuid', 'details']

    def __init__(self,
                 id,  # pylint:disable=redefined-builtin
                 uuid,
                 experiment_group,
                 created_at,
                 status,
                 message=None,
                 details=None):
        self.id = id
        self.uuid = uuid
        self.experiment_group = experiment_group
        self.created_at = self.localize_date(created_at)
        self.status = status
        self.message = message
        self.details = details


class ProjectSchema(Schema):
    id = fields.Int(allow_none=True)
    name = fields.Str(validate=validate.Regexp(regex=r'^[-a-zA-Z0-9_]+\Z'))
    user = fields.Str(validate=validate.Regexp(regex=r'^[-a-zA-Z0-9_]+\Z'), allow_none=True)
    unique_name = fields.Str(allow_none=True)
    uuid = UUID(allow_none=True)
    description = fields.Str(allow_none=True)
    tags = fields.List(fields.Str(), allow_none=True)
    is_public = fields.Boolean(allow_none=True)
    has_code = fields.Bool(allow_none=True)
    created_at = fields.LocalDateTime(allow_none=True)
    updated_at = fields.LocalDateTime(allow_none=True)
    num_experiments = fields.Int(allow_none=True)
    num_independent_experiments = fields.Int(allow_none=True)
    num_experiment_groups = fields.Int(allow_none=True)
    num_jobs = fields.Int(allow_none=True)
    num_builds = fields.Int(allow_none=True)
    has_tensorboard = fields.Bool(allow_none=True)
    has_notebook = fields.Bool(allow_none=True)
    experiment_groups = fields.Nested(ExperimentGroupSchema, many=True, allow_none=True)
    experiments = fields.Nested(ExperimentSchema, many=True, allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ProjectConfig(**data)

    @post_dump
    def unmake(self, data):
        return ProjectConfig.remove_reduced_attrs(data)


class ProjectConfig(BaseConfig):
    SCHEMA = ProjectSchema
    IDENTIFIER = 'project'
    DEFAULT_EXCLUDE_ATTRIBUTES = [
        'id', 'uuid', 'description', 'updated_at',
        'experiment_groups', 'experiments', 'has_code', 'user'
    ]
    DATETIME_ATTRIBUTES = ['created_at', 'updated_at']

    def __init__(self,
                 name,
                 id=None,  # pylint:disable=redefined-builtin
                 user=None,
                 unique_name=None,
                 uuid=None,
                 description=None,
                 is_public=True,
                 tags=None,
                 has_code=False,
                 has_tensorboard=False,
                 has_notebook=False,
                 created_at=None,
                 updated_at=None,
                 num_experiments=0,
                 num_experiment_groups=0,
                 num_independent_experiments=0,
                 num_jobs=0,
                 num_builds=0,
                 experiments=None,
                 experiment_groups=None):
        self.name = name
        self.id = id
        self.user = user
        self.unique_name = unique_name
        self.uuid = uuid
        self.description = description
        self.is_public = is_public
        self.tags = tags
        self.has_code = has_code
        self.has_tensorboard = has_tensorboard
        self.has_notebook = has_notebook
        self.created_at = self.localize_date(created_at)
        self.updated_at = self.localize_date(updated_at)
        self.num_experiments = num_experiments
        self.num_independent_experiments = num_independent_experiments
        self.num_experiment_groups = num_experiment_groups
        self.num_jobs = num_jobs
        self.num_builds = num_builds
        self.experiments = experiments
        self.experiment_groups = experiment_groups



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.layers.embeddings import EmbeddingConfig
from polyaxon_schemas.layers.wrappers import (
    BidirectionalConfig,
    TimeDistributedConfig,
    WrapperConfig
)


class TestWrapperConfigs(TestCase):
    @staticmethod
    def assert_wrapper_config(wrapper_class):
        config_dict = {
            'layer': EmbeddingConfig(input_dim=30, output_dim=30).to_schema(),
        }
        config = wrapper_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_wrapper_config(self):
        self.assert_wrapper_config(WrapperConfig)
        self.assert_wrapper_config(TimeDistributedConfig)
        self.assert_wrapper_config(BidirectionalConfig)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.initializations import ZerosInitializerConfig
from polyaxon_schemas.layers.advanced_activations import (
    ELUConfig,
    LeakyReLUConfig,
    PReLUConfig,
    ThresholdedReLUConfig
)


class TestAdvancedActivationConfigs(TestCase):
    def test_leaky_relu_config(self):
        config_dict = {
            'alpha': 0.3,
        }
        config = LeakyReLUConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_prelu_config(self):
        config_dict = {
            'alpha_initializer': ZerosInitializerConfig().to_schema(),
            'alpha_regularizer': None,
            'alpha_constraint': None,
            'shared_axes': None
        }
        config = PReLUConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_elu_config(self):
        config_dict = {
            'alpha': 0.1
        }
        config = ELUConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_thresholded_relu_config(self):
        config_dict = {
            'theta': 0.1
        }
        config = ThresholdedReLUConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.constraints import MaxNormConfig, MinMaxNormConfig, UnitNormConfig


class TestConstraintConfigs(TestCase):
    def test_max_norm_constraint_config(self):
        config_dict = {
            'max_value': 2,
            'axis': 1,
        }
        config = MaxNormConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_unit_norm_constraint_config(self):
        config_dict = {
            'axis': 1,
        }
        config = UnitNormConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_min_max_norm_constraint_config(self):
        config_dict = {
            'min_value': 1,
            'max_value': 2,
            'rate': 1.,
            'axis': 1,
        }
        config = MinMaxNormConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, post_dump, post_load, validate

from polyaxon_schemas.layers.base import BaseLayerConfig, BaseLayerSchema


class GaussianNoiseSchema(BaseLayerSchema):
    stddev = fields.Float()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return GaussianNoiseConfig(**data)

    @post_dump
    def unmake(self, data):
        return GaussianNoiseConfig.remove_reduced_attrs(data)


class GaussianNoiseConfig(BaseLayerConfig):
    """Apply additive zero-centered Gaussian noise.

    This is useful to mitigate overfitting
    (you could see it as a form of random data augmentation).
    Gaussian Noise (GS) is a natural choice as corruption process
    for real valued inputs.

    As it is a regularization layer, it is only active at training time.

    Args:
        stddev: float, standard deviation of the noise distribution.

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same shape as input.

    Polyaxonfile usage:

    ```yaml
    GaussianNoise:
      stddev: 0.5
    ```
    """
    IDENTIFIER = 'GaussianNoise'
    SCHEMA = GaussianNoiseSchema

    def __init__(self, stddev, **kwargs):
        super(GaussianNoiseConfig, self).__init__(**kwargs)
        self.stddev = stddev


class GaussianDropoutSchema(BaseLayerSchema):
    rate = fields.Float(validate=validate.Range(0, 1))

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return GaussianDropoutConfig(**data)

    @post_dump
    def unmake(self, data):
        return GaussianDropoutConfig.remove_reduced_attrs(data)


class GaussianDropoutConfig(BaseLayerConfig):
    """Apply multiplicative 1-centered Gaussian noise.

    As it is a regularization layer, it is only active at training time.

    Args:
        rate: float, drop probability (as with `Dropout`).
            The multiplicative noise will have
            standard deviation `sqrt(rate / (1 - rate))`.

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same shape as input.

    References:
        - [Dropout: A Simple Way to Prevent Neural Networks from Overfitting
          Srivastava, Hinton, et al.
          2014](http://www.cs.toronto.edu/~rsalakhu/papers/srivastava14a.pdf)

    Polyaxonfile usage:

    ```yaml
    GaussianDropout:
      rate: 0.7
    ```
    """
    IDENTIFIER = 'GaussianDropout'
    SCHEMA = GaussianDropoutSchema

    def __init__(self, rate, **kwargs):
        super(GaussianDropoutConfig, self).__init__(**kwargs)
        self.rate = rate


class AlphaDropoutSchema(BaseLayerSchema):
    rate = fields.Float(validate=validate.Range(0, 1))
    noise_shape = fields.List(fields.Int(), default=None, missing=None)
    seed = fields.Int(default=None, missing=None)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return AlphaDropoutConfig(**data)

    @post_dump
    def unmake(self, data):
        return AlphaDropoutConfig.remove_reduced_attrs(data)


class AlphaDropoutConfig(BaseLayerConfig):
    """Applies Alpha Dropout to the input.

    Alpha Dropout is a `Dropout` that keeps mean and variance of inputs
    to their original values, in order to ensure the self-normalizing property
    even after this dropout.
    Alpha Dropout fits well to Scaled Exponential Linear Units
    by randomly setting activations to the negative saturation value.

    Args:
        rate: float, drop probability (as with `Dropout`).
            The multiplicative noise will have
            standard deviation `sqrt(rate / (1 - rate))`.
        seed: A Python integer to use as random seed.

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same shape as input.

    References:
        - [Self-Normalizing Neural Networks](https://arxiv.org/abs/1706.02515)

    Polyaxonfile usage:

    ```yaml
    AlphaDropout:
      rate: 0.7
    ```
    """
    IDENTIFIER = 'AlphaDropout'
    SCHEMA = AlphaDropoutSchema

    def __init__(self, rate, noise_shape=None, seed=None, **kwargs):
        super(AlphaDropoutConfig, self).__init__(**kwargs)
        self.rate = rate
        self.noise_shape = noise_shape
        self.seed = seed



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import ast
import numpy as np
import six

from collections import Mapping
from datetime import datetime
from pytz import timezone

from marshmallow import ValidationError, fields, post_dump, post_load, validate
from marshmallow.base import FieldABC
from marshmallow.utils import _Missing


def get_obj_or_list_obj(container, value, min_length=None, max_length=None):
    try:
        return container.deserialize(value)
    except (ValueError, TypeError, ValidationError):
        pass

    if not isinstance(value, (list, tuple)):
        raise ValidationError("This field expects an int or a list of ints.")

    value = validate.Length(min=min_length, max=max_length)(value)
    try:
        return [container.deserialize(v) for v in value]
    except (ValueError, TypeError):
        raise ValidationError("This field expects an int or a list of ints.")


class UUID(fields.UUID):
    """A UUID field."""

    def _serialize(self, value, attr, obj):
        validated = str(self._validated(value).hex) if value is not None else None
        return super(fields.String, self)._serialize(validated, attr, obj)  # noqa


class ObjectOrListObject(fields.Field):
    def __init__(self, cls_or_instance, min=None, max=None, **kwargs):  # noqa
        self.min = min
        self.max = max

        super(ObjectOrListObject, self).__init__(**kwargs)
        if isinstance(cls_or_instance, type):
            if not issubclass(cls_or_instance, FieldABC):
                raise ValueError('The type of the list elements '
                                 'must be a subclass of '
                                 'marshmallow.base.FieldABC')
            self.container = cls_or_instance()
        else:
            if not isinstance(cls_or_instance, FieldABC):
                raise ValueError('The instances of the list '
                                 'elements must be of type '
                                 'marshmallow.base.FieldABC')
            self.container = cls_or_instance

    def _deserialize(self, value, attr, data):
        return get_obj_or_list_obj(self.container, value, self.min, self.max)


class Tensor(fields.Field):
    def _deserialize(self, value, attr, data):
        if isinstance(value, six.string_types):
            return [value, 0, 0]
        if isinstance(value, list) and len(value) == 3:
            condition = (isinstance(value[0], str) and
                         isinstance(value[1], int) and
                         isinstance(value[1], int))
            if condition:
                return value
        raise ValidationError("This field expects a str or a list of [str, int, int].")


class PValue(fields.Field):
    def _deserialize(self, value, attr, data):
        if isinstance(value, (list, tuple)) and len(value) == 2:
            if isinstance(value[1], float) and 0 <= value[1] < 1:
                return value
        raise ValidationError("This field expects a list of [value<Any>, dist<float>].")


class IntOrStr(fields.Str):
    default_error_messages = {
        'invalid': 'Not a valid string or int.',
        'invalid_utf8': 'Not a valid utf-8 string.'
    }

    def _serialize(self, value, attr, obj):
        if isinstance(value, int):
            return int(value)

        return super(IntOrStr, self)._serialize(value=value, attr=attr, obj=obj)

    def _deserialize(self, value, attr, data):
        if isinstance(value, int):
            return int(value)

        return super(IntOrStr, self)._deserialize(value=value, attr=attr, data=data)


class Range(fields.Field):
    REQUIRED_KEYS = ['start', 'stop', 'step']
    OPTIONAL_KEY = None
    KEYS = REQUIRED_KEYS

    def _deserialize(self, value, attr, data):
        if isinstance(value, six.string_types):
            value = value.split(':')
        elif isinstance(value, Mapping):
            if set(self.REQUIRED_KEYS) - set(six.iterkeys(value)):
                raise ValidationError("{} dict must have {} keys {}.".format(
                    self.__class__.__name__, len(self.REQUIRED_KEYS), self.REQUIRED_KEYS))
            if len(value) == len(self.REQUIRED_KEYS):
                value = [value[k] for k in self.REQUIRED_KEYS]
            elif len(value) == len(self.KEYS):
                value = [value[k] for k in self.KEYS]
        elif not isinstance(value, list):
            raise ValidationError(
                "{} accept values formatted as the following:\n"
                " * str: {}\n"
                " * dict: {}\n"
                " * list: {}".format(
                    self.__class__.__name__,
                    ':'.join(self.REQUIRED_KEYS),
                    dict(zip(self.REQUIRED_KEYS,
                             ['v{}'.format(i) for i in range(len(self.REQUIRED_KEYS))])),
                    self.REQUIRED_KEYS))

        if len(value) != len(self.REQUIRED_KEYS) and len(value) != len(self.KEYS):
            raise ValidationError("{} requires {} or {} elements received {}".format(
                self.__class__.__name__,
                len(self.REQUIRED_KEYS),
                len(self.KEYS),
                len(value)))

        for i, v in enumerate(value):
            try:
                float(v)
            except (ValueError, TypeError):
                raise ValidationError(
                    "{}: {} must of type int or float, received instead {}".format(
                        self.__class__.__name__, self.REQUIRED_KEYS[i], v
                    ))
            if not isinstance(v, (int, float)):
                value[i] = ast.literal_eval(v)

        # Check that lower value is  smaller than higher value
        if value[0] >= value[1]:
            raise ValidationError(
                "{key2} value must strictly higher that {key1} value, "
                "received instead {key1}: {val1}, {key2}: {val2}".format(
                    key1=self.REQUIRED_KEYS[0],
                    key2=self.REQUIRED_KEYS[1],
                    val1=value[0],
                    val2=value[1]))
        if len(self.REQUIRED_KEYS) == 3 and value[2] == 0:
            raise ValidationError("{} cannot be 0".format(self.REQUIRED_KEYS[2]))

        value = dict(zip(self.KEYS, value))
        return value


class LinSpace(Range):
    REQUIRED_KEYS = ['start', 'stop', 'num']
    KEYS = REQUIRED_KEYS


class GeomSpace(Range):
    REQUIRED_KEYS = ['start', 'stop', 'num']
    KEYS = REQUIRED_KEYS


class LogSpace(Range):
    REQUIRED_KEYS = ['start', 'stop', 'num']
    OPTIONAL_KEYS = ['base']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


def uniform(low, high, size=None, rand_generator=None):
    rand_generator = rand_generator or np.random
    return rand_generator.uniform(low=low, high=high, size=size)


def quniform(low, high, q, size=None, rand_generator=None):
    value = uniform(low=low, high=high, size=size, rand_generator=rand_generator)
    return np.round(value // q) * q


def loguniform(low, high, size=None, rand_generator=None):
    value = uniform(low=low, high=high, size=size, rand_generator=rand_generator)
    return np.exp(value)


def qloguniform(low, high, q, size=None, rand_generator=None):
    value = loguniform(low=low, high=high, size=size, rand_generator=rand_generator)
    return np.round(value // q) * q


def normal(loc, scale, size=None, rand_generator=None):
    rand_generator = rand_generator or np.random
    return rand_generator.normal(loc=loc, scale=scale, size=size)


def qnormal(loc, scale, q, size=None, rand_generator=None):
    draw = normal(loc=loc, scale=scale, size=size, rand_generator=rand_generator)
    return np.round(draw // q) * q


def lognormal(loc, scale, size=None, rand_generator=None):
    rand_generator = rand_generator or np.random
    return rand_generator.lognormal(mean=loc, sigma=scale, size=size)


def qlognormal(loc, scale, q, size=None, rand_generator=None):
    draw = lognormal(loc=loc, scale=scale, size=size, rand_generator=rand_generator)
    return np.exp(draw)


def validate_pvalues(values):
    dists = [v for v in values if v]
    if sum(dists) > 1:
        raise ValidationError('The distribution of different outcomes should sum to 1.')


def pvalues(values, size=None, rand_generator=None):
    rand_generator = rand_generator or np.random
    keys = [v[0] for v in values]
    dists = [v[1] for v in values]
    validate_pvalues(dists)
    indices = rand_generator.multinomial(1, dists, size=size)
    if size is None:
        return keys[indices.argmax()]
    return [keys[ind.argmax()] for ind in indices]


class Uniform(Range):
    REQUIRED_KEYS = ['low', 'high']
    OPTIONAL_KEYS = ['size']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


class QUniform(Range):
    REQUIRED_KEYS = ['low', 'high', 'q']
    OPTIONAL_KEYS = ['size']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


class LogUniform(Range):
    REQUIRED_KEYS = ['low', 'high']
    OPTIONAL_KEYS = ['size']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


class QLogUniform(Range):
    REQUIRED_KEYS = ['low', 'high', 'q']
    OPTIONAL_KEYS = ['size']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


class Normal(Range):
    REQUIRED_KEYS = ['loc', 'scale']
    OPTIONAL_KEYS = ['size']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


class QNormal(Range):
    REQUIRED_KEYS = ['loc', 'scale', 'q']
    OPTIONAL_KEYS = ['size']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


class LogNormal(Range):
    REQUIRED_KEYS = ['loc', 'scale']
    OPTIONAL_KEYS = ['size']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


class QLogNormal(Range):
    REQUIRED_KEYS = ['loc', 'scale', 'q']
    OPTIONAL_KEYS = ['size']
    KEYS = REQUIRED_KEYS + OPTIONAL_KEYS


class StrOrFct(fields.Str):
    def serialize(self, attr, obj, accessor=None):
        value = getattr(obj, attr)
        if hasattr(value, '__call__') and hasattr(value, '__name__'):
            return value.__name__

        return super(StrOrFct, self).serialize(attr, obj, accessor)


class DType(fields.Str):
    def serialize(self, attr, obj, accessor=None):
        value = getattr(obj, attr)
        if hasattr(value, 'name'):
            return value.name

        return super(DType, self).serialize(attr, obj, accessor)


class UnknownSchemaMixin(object):
    @post_load(pass_original=True)
    def _handle_load_unknown(self, data, original):
        """Preserve unknown keys during deserialization."""
        for key, val in original.items():
            if key not in self.fields:
                data[key] = val
        return data

    @post_dump(pass_original=True)
    def _handle_dump_unknown(self, data, original):
        """Preserve unknown keys during deserialization."""
        for key, val in original.items():
            if key not in self.fields:
                data[key] = val
        return data


def to_list(value):
    if isinstance(value, (np.ndarray, list, tuple)):
        return list(value)
    return [value]


missing = _Missing()


def get_value(key, obj, default=missing):
    """Helper for pulling a keyed value off various types of objects"""
    if isinstance(key, int):
        return _get_value_for_key(key, obj, default)
    return _get_value_for_keys(key.split('.'), obj, default)


def _get_value_for_keys(keys, obj, default):
    if len(keys) == 1:
        return _get_value_for_key(keys[0], obj, default)
    return _get_value_for_keys(keys[1:], _get_value_for_key(keys[0], obj, default), default)


def _get_value_for_key(key, obj, default):
    try:
        return obj[key]
    except (KeyError, AttributeError, IndexError, TypeError):
        try:
            attr = getattr(obj, key)
            if callable(attr):
                if hasattr(attr, 'get_config') or hasattr(attr, 'SCHEMA'):
                    return attr

                return attr.__name__

            return attr
        except AttributeError:
            return default


def to_camel_case(snake_str):
    split_str = snake_str.split('_')
    if len(split_str) == 1:
        try:
            return snake_str if str.isupper(snake_str[0]) else snake_str.title()
        except TypeError:
            return snake_str if six.text_type.isupper(snake_str[0]) else snake_str.title()
    return "".join(x.title() for x in split_str)


ACTIVATION_VALUES = [
    'softmax', 'elu', 'selu', 'softplus', 'softsign', 'relu', 'tanh', 'sigmoid',
    'hard_sigmoid', 'linear'
]
TIME_ZONE = timezone('UTC')


class SearchAlgorithms(object):
    GRID = 'grid'
    RANDOM = 'random'
    HYPERBAND = 'hyperband'
    BO = 'bo'  # bayesian optimization

    GRID_VALUES = [GRID, GRID.upper(), GRID.capitalize()]
    RANDOM_VALUES = [RANDOM, RANDOM.upper(), RANDOM.capitalize()]
    HYPERBAND_VALUES = [HYPERBAND, HYPERBAND.upper(), HYPERBAND.capitalize()]
    BO_VALUES = [BO, BO.upper(), BO.capitalize()]

    VALUES = GRID_VALUES + RANDOM_VALUES + HYPERBAND_VALUES + BO_VALUES

    @classmethod
    def is_grid(cls, value):
        return value in cls.GRID_VALUES

    @classmethod
    def is_random(cls, value):
        return value in cls.RANDOM_VALUES

    @classmethod
    def is_hyperband(cls, value):
        return value in cls.HYPERBAND_VALUES

    @classmethod
    def is_bo(cls, value):
        return value in cls.BO_VALUES


class AcquisitionFunctions(object):
    UCB = 'ucb'
    EI = 'ei'
    POI = 'poi'

    UCB_VALUES = [UCB, UCB.upper(), UCB.capitalize()]
    EI_VALUES = [EI, EI.upper(), EI.capitalize()]
    POI_VALUES = [POI, POI.upper(), POI.capitalize()]

    VALUES = UCB_VALUES + EI_VALUES + POI_VALUES

    @classmethod
    def is_ucb(cls, value):
        return value in cls.UCB_VALUES

    @classmethod
    def is_ei(cls, value):
        return value in cls.EI_VALUES

    @classmethod
    def is_poi(cls, value):
        return value in cls.POI_VALUES


class GaussianProcessesKernels(object):
    RBF = 'rbf'
    MATERN = 'matern'

    RBF_VALUES = [RBF, RBF.upper(), RBF.capitalize()]
    MATERN_VALUES = [MATERN, MATERN.upper(), MATERN.capitalize()]

    VALUES = RBF_VALUES + MATERN_VALUES

    @classmethod
    def is_rbf(cls, value):
        return value in cls.RBF_VALUES

    @classmethod
    def is_mattern(cls, value):
        return value in cls.MATERN_VALUES


class Optimization(object):
    MAXIMIZE = 'maximize'
    MINIMIZE = 'minimize'

    MAXIMIZE_VALUES = [MAXIMIZE, MAXIMIZE.upper(), MAXIMIZE.capitalize()]
    MINIMIZE_VALUES = [MINIMIZE, MINIMIZE.upper(), MINIMIZE.capitalize()]

    VALUES = MAXIMIZE_VALUES + MINIMIZE_VALUES

    @classmethod
    def maximize(cls, value):
        return value in cls.MAXIMIZE_VALUES

    @classmethod
    def minimize(cls, value):
        return value in cls.MINIMIZE_VALUES


class ResourceTypes(object):
    INT = 'int'
    FLOAT = 'float'

    INT_VALUES = [INT, INT.upper(), INT.capitalize()]
    FLOAT_VALUES = [FLOAT, FLOAT.upper(), FLOAT.capitalize()]

    VALUES = INT_VALUES + FLOAT_VALUES

    @classmethod
    def is_int(cls, value):
        return value in cls.INT_VALUES

    @classmethod
    def is_float(cls, value):
        return value in cls.FLOAT_VALUES


class EarlyStoppingPolicy(object):
    EXPERIMENT = 'experiment'
    GROUP = 'group'
    ALL = 'all'

    EXPERIMENT_VALUES = [EXPERIMENT, EXPERIMENT.upper(), EXPERIMENT.capitalize()]
    GROUP_VALUES = [GROUP, GROUP.upper(), GROUP.capitalize()]
    ALL_VALUES = [ALL, ALL.upper(), ALL.capitalize()]

    VALUES = EXPERIMENT_VALUES + GROUP_VALUES + ALL_VALUES

    @classmethod
    def stop_experiment(cls, value):
        return value in cls.EXPERIMENT_VALUES

    @classmethod
    def stop_group(cls, value):
        return value in cls.GROUP_VALUES

    @classmethod
    def stop_all(cls, value):
        return value in cls.ALL_VALUES


class TaskType(object):
    MASTER = 'master'
    PS = 'ps'
    WORKER = 'worker'
    SERVER = 'server'
    SCHEDULER = 'scheduler'

    VALUES = [MASTER, PS, WORKER, SERVER, SCHEDULER]


class Frameworks(object):
    TENSORFLOW = 'tensorflow'
    MXNET = 'mxnet'
    HOROVOD = 'horovod'
    PYTORCH = 'pytorch'

    VALUES = [TENSORFLOW, MXNET, HOROVOD, PYTORCH]


def local_now():
    return TIME_ZONE.localize(datetime.utcnow()).replace(microsecond=0)


def humanize_timesince(start_time):
    """Creates a string representation of time since the given `start_time`."""
    if not start_time:
        return start_time

    delta = local_now() - start_time

    # assumption: negative delta values originate from clock
    #             differences on different app server machines
    if delta.total_seconds() < 0:
        return 'a few seconds ago'

    num_years = delta.days // 365
    if num_years > 0:
        return '{} year{} ago'.format(
            *((num_years, 's') if num_years > 1 else (num_years, '')))

    num_weeks = delta.days // 7
    if num_weeks > 0:
        return '{} week{} ago'.format(
            *((num_weeks, 's') if num_weeks > 1 else (num_weeks, '')))

    num_days = delta.days
    if num_days > 0:
        return '{} day{} ago'.format(
            *((num_days, 's') if num_days > 1 else (num_days, '')))

    num_hours = delta.seconds // 3600
    if num_hours > 0:
        return '{} hour{} ago'.format(*((num_hours, 's') if num_hours > 1 else (num_hours, '')))

    num_minutes = delta.seconds // 60
    if num_minutes > 0:
        return '{} minute{} ago'.format(
            *((num_minutes, 's') if num_minutes > 1 else (num_minutes, '')))

    return 'a few seconds ago'


def humanize_timedelta(seconds):
    """Creates a string representation of timedelta."""
    hours, remainder = divmod(seconds, 3600)
    days, hours = divmod(hours, 24)
    minutes, seconds = divmod(remainder, 60)

    if days:
        result = '{}d'.format(days)
        if hours:
            result += ' {}h'.format(hours)
        if minutes:
            result += ' {}m'.format(minutes)
        return result

    if hours:
        result = '{}h'.format(hours)
        if minutes:
            result += ' {}m'.format(minutes)
        return result

    if minutes:
        result = '{}m'.format(minutes)
        if seconds:
            result += ' {}s'.format(seconds)
        return result

    return '{}s'.format(seconds)


def to_percentage(number, rounding=2):
    """Creates a percentage string representation from the given `number`. The
    number is multiplied by 100 before adding a '%' character.

    Raises `ValueError` if `number` cannot be converted to a number.
    """
    number = float(number) * 100
    number_as_int = int(number)
    rounded = round(number, rounding)

    return '{}%'.format(number_as_int if number_as_int == rounded else rounded)


def to_unit_memory(number):
    """Creates a string representation of memory size given `number`."""
    kb = 1024

    number /= kb

    if number < 100:
        return '{} Kb'.format(round(number, 2))

    number /= kb
    if number < 300:
        return '{} Mb'.format(round(number, 2))

    number /= kb

    return '{} Gb'.format(round(number, 2))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema
from polyaxon_schemas.processing.feature_processors import FeatureProcessorsSchema


class BasePipelineSchema(Schema):
    name = fields.Str(allow_none=True)
    feature_processors = fields.Nested(FeatureProcessorsSchema, allow_none=True)
    shuffle = fields.Bool(allow_none=True)
    num_epochs = fields.Int(allow_none=True)
    batch_size = fields.Int(allow_none=True)
    bucket_boundaries = fields.List(fields.Int(), allow_none=True)
    allow_smaller_final_batch = fields.Bool(allow_none=True)
    dynamic_pad = fields.Bool(allow_none=True)
    min_after_dequeue = fields.Int(allow_none=True)
    num_threads = fields.Int(allow_none=True)
    capacity = fields.Int(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return BasePipelineConfig(**data)

    @post_dump
    def unmake(self, data):
        return BasePipelineConfig.remove_reduced_attrs(data)


class BasePipelineConfig(BaseConfig):
    """Abstract InputPipeline class. All input pipelines must inherit from this.
    An InputPipeline defines how data is read, parsed, and separated into
    features and labels.

    Args:
        name: `str`, name to give for this pipeline.
        feature_processors: `dict`, list of modules to call for each feature to be processed.
        shuffle: `bool`, If true, shuffle the data.
        num_epochs: `int`, Number of times to iterate through the dataset. If None, iterate forever.
        batch_size: The new batch size pulled from the queue (all queues will have the same size).
            If a list is passed in then each bucket will have a different batch_size.
            (python int, int32 scalar or iterable of integers of length num_buckets).
        bucket_boundaries: `list` of `int` or `None`, increasing non-negative numbers.
            The edges of the buckets to use when bucketing tensors.
            Two extra buckets are created, one for input_length < bucket_boundaries[0]
            and one for input_length >= bucket_boundaries[-1].
        allow_smaller_final_batch: `bool`, whether to allow a last small batch.
        dynamic_pad: `bool`, Allow variable dimensions in input shapes.
            The given dimensions are padded upon dequeue so that tensors
            within a batch have the same shapes.
        min_after_dequeue: `int`.
        num_threads: `int`. The number of threads enqueuing tensors.
        capacity: `int`, The maximum number of minibatches in the top queue,
            and also the maximum number of elements within each bucket.
    """
    IDENTIFIER = 'BasePipeline'
    SCHEMA = BasePipelineSchema
    REDUCED_ATTRIBUTES = ['feature_processors']

    def __init__(self,
                 name='Pipeline',
                 feature_processors=None,
                 shuffle=True,
                 num_epochs=1,
                 batch_size=64,
                 bucket_boundaries=None,
                 allow_smaller_final_batch=True,
                 dynamic_pad=False,
                 min_after_dequeue=5000,
                 num_threads=3,
                 capacity=None):
        self.name = name
        self.feature_processors = feature_processors
        self.shuffle = shuffle
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.bucket_boundaries = bucket_boundaries
        self.allow_smaller_final_batch = allow_smaller_final_batch
        self.dynamic_pad = dynamic_pad
        self.min_after_dequeue = min_after_dequeue
        self.num_threads = num_threads
        self.capacity = capacity or min_after_dequeue + num_threads * batch_size


class TFRecordImagePipelineSchema(BasePipelineSchema):
    data_files = fields.List(fields.Str(), allow_none=True)
    meta_data_file = fields.Str()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return TFRecordImagePipelineConfig(**data)

    @post_dump
    def unmake(self, data):
        return TFRecordImagePipelineConfig.remove_reduced_attrs(data)


class TFRecordImagePipelineConfig(BasePipelineConfig):
    """A Pipeline to convert TF-Records to images.

    Args:
        name: `str`, name to give for this pipeline.
        feature_processors: `dict`, list of modules to call for each feature to be processed.
        shuffle: `bool`, If true, shuffle the data.
        num_epochs: `int`, Number of times to iterate through the dataset. If None, iterate forever.
        batch_size: The new batch size pulled from the queue (all queues will have the same size).
            If a list is passed in then each bucket will have a different batch_size.
            (python int, int32 scalar or iterable of integers of length num_buckets).
        bucket_boundaries: `list` of `int` or `None`, increasing non-negative numbers.
            The edges of the buckets to use when bucketing tensors.
            Two extra buckets are created, one for input_length < bucket_boundaries[0]
            and one for input_length >= bucket_boundaries[-1].
        allow_smaller_final_batch: `bool`, whether to allow a last small batch.
        dynamic_pad: `bool`, Allow variable dimensions in input shapes.
            The given dimensions are padded upon dequeue so that tensors
            within a batch have the same shapes.
        min_after_dequeue: `int`.
        num_threads: `int`. The number of threads enqueuing tensors.
        capacity: `int`, The maximum number of minibatches in the top queue,
            and also the maximum number of elements within each bucket.
        data_files: `list` of `str`. List of the filenames for data.
        meta_data_file: `str`. Metadata filename

    Polyaxonfile usage:

    ```yaml
    TFRecordImagePipeline:
      batch_size: 64
      num_epochs: 1
      shuffle: true
      dynamic_pad: false
      data_files: ["../data/mnist/mnist_train.tfrecord"]
      meta_data_file: "../data/mnist/meta_data.json"
      feature_processors:
        image:
          input_layers: [image]
          layers:
            - Cast:
                dtype: float32
    ```
    """
    IDENTIFIER = 'TFRecordImagePipeline'
    SCHEMA = TFRecordImagePipelineSchema

    def __init__(self, data_files, meta_data_file, name='TFRecordImagePipeline', **kwargs):
        super(TFRecordImagePipelineConfig, self).__init__(name=name, **kwargs)
        self.data_files = data_files
        self.meta_data_file = meta_data_file


class TFRecordSequencePipelineSchema(BasePipelineSchema):
    data_files = fields.List(fields.Str(), allow_none=True)
    meta_data_file = fields.Str()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return TFRecordSequencePipelineConfig(**data)

    @post_dump
    def unmake(self, data):
        return TFRecordSequencePipelineConfig.remove_reduced_attrs(data)


class TFRecordSequencePipelineConfig(BasePipelineConfig):
    """A Pipeline to convert TF-Records to sequences.

    At least one sequence must be `source_token`.

    Args:
        name: `str`, name to give for this pipeline.
        feature_processors: `dict`, list of modules to call for each feature to be processed.
        shuffle: `bool`, If true, shuffle the data.
        num_epochs: `int`, Number of times to iterate through the dataset. If None, iterate forever.
        batch_size: The new batch size pulled from the queue (all queues will have the same size).
            If a list is passed in then each bucket will have a different batch_size.
            (python int, int32 scalar or iterable of integers of length num_buckets).
        bucket_boundaries: `list` of `int` or `None`, increasing non-negative numbers.
            The edges of the buckets to use when bucketing tensors.
            Two extra buckets are created, one for input_length < bucket_boundaries[0]
            and one for input_length >= bucket_boundaries[-1].
        allow_smaller_final_batch: `bool`, whether to allow a last small batch.
        dynamic_pad: `bool`, Allow variable dimensions in input shapes.
            The given dimensions are padded upon dequeue so that tensors
            within a batch have the same shapes.
        min_after_dequeue: `int`.
        num_threads: `int`. The number of threads enqueuing tensors.
        capacity: `int`, The maximum number of minibatches in the top queue,
            and also the maximum number of elements within each bucket.
        data_files: `list` of `str`. List of the filenames for data.
        meta_data_file: `str`. Metadata filename

    Polyaxonfile usage:

    ```yaml
    TFRecordSequencePipeline:
      batch_size: 64
      num_epochs: 1
      shuffle: true
      dynamic_pad: false
      data_files: ["data.tfrecord"]
      meta_data_file: "meta_data.json"
      feature_processors:
        image:
          input_layers: [sequence]
          layers:
            - Cast:
                dtype: float32
    ```
    """
    IDENTIFIER = 'TFRecordSequencePipeline'
    SCHEMA = TFRecordSequencePipelineSchema

    def __init__(self, data_files, meta_data_file, name='TFRecordSequencePipeline', **kwargs):
        super(TFRecordSequencePipelineConfig, self).__init__(name=name, **kwargs)
        self.data_files = data_files
        self.meta_data_file = meta_data_file


class ParallelTextPipelineSchema(BasePipelineSchema):
    source_files = fields.List(fields.Str(), allow_none=True)
    target_files = fields.List(fields.Str(), allow_none=True)
    source_delimiter = fields.Str(allow_none=True)
    target_delimiter = fields.Str(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ParallelTextPipelineConfig(**data)

    @post_dump
    def unmake(self, data):
        return ParallelTextPipelineConfig.remove_reduced_attrs(data)


class ParallelTextPipelineConfig(BasePipelineConfig):
    """An input pipeline that reads two parallel (line-by-line aligned) text files.

    Args:
        name: `str`, name to give for this pipeline.
        feature_processors: `dict`, list of modules to call for each feature to be processed.
        shuffle: `bool`, If true, shuffle the data.
        num_epochs: `int`, Number of times to iterate through the dataset. If None, iterate forever.
        batch_size: The new batch size pulled from the queue (all queues will have the same size).
            If a list is passed in then each bucket will have a different batch_size.
            (python int, int32 scalar or iterable of integers of length num_buckets).
        bucket_boundaries: `list` of `int` or `None`, increasing non-negative numbers.
            The edges of the buckets to use when bucketing tensors.
            Two extra buckets are created, one for input_length < bucket_boundaries[0]
            and one for input_length >= bucket_boundaries[-1].
        allow_smaller_final_batch: `bool`, whether to allow a last small batch.
        dynamic_pad: `bool`, Allow variable dimensions in input shapes.
            The given dimensions are padded upon dequeue so that tensors
            within a batch have the same shapes.
        min_after_dequeue: `int`.
        num_threads: `int`. The number of threads enqueuing tensors.
        capacity: `int`, The maximum number of minibatches in the top queue,
            and also the maximum number of elements within each bucket.
        source_files: An array of file names for the source data.
        target_files: An array of file names for the target data. These must
          be aligned to the `source_files`.
        source_delimiter: A character to split the source text on. Defaults
          to  " " (space). For character-level training this can be set to the
          empty string.
        target_delimiter: Same as `source_delimiter` but for the target text.
    """
    IDENTIFIER = 'ParallelTextPipeline'
    SCHEMA = ParallelTextPipelineSchema

    def __init__(self,
                 source_files=None,
                 target_files=None,
                 source_delimiter="",
                 target_delimiter="",
                 name='ParallelTextPipeline',
                 **kwargs):
        super(ParallelTextPipelineConfig, self).__init__(name=name, **kwargs)
        self.source_files = source_files
        self.target_files = target_files
        self.source_delimiter = source_delimiter
        self.target_delimiter = target_delimiter


class TFRecordSourceSequencePipelineSchema(BasePipelineSchema):
    files = fields.List(fields.Str(), allow_none=True)
    source_field = fields.Str(allow_none=True)
    target_field = fields.Str(allow_none=True)
    source_delimiter = fields.Str(allow_none=True)
    target_delimiter = fields.Str(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return TFRecordSourceSequencePipelineConfig(**data)

    @post_dump
    def unmake(self, data):
        return TFRecordSourceSequencePipelineConfig.remove_reduced_attrs(data)


class TFRecordSourceSequencePipelineConfig(BasePipelineConfig):
    """An input pipeline that reads a TFRecords containing both source and target sequences.

    Args:
        name: `str`, name to give for this pipeline.
        feature_processors: `dict`, list of modules to call for each feature to be processed.
        shuffle: `bool`, If true, shuffle the data.
        num_epochs: `int`, Number of times to iterate through the dataset. If None, iterate forever.
        batch_size: The new batch size pulled from the queue (all queues will have the same size).
            If a list is passed in then each bucket will have a different batch_size.
            (python int, int32 scalar or iterable of integers of length num_buckets).
        bucket_boundaries: `list` of `int` or `None`, increasing non-negative numbers.
            The edges of the buckets to use when bucketing tensors.
            Two extra buckets are created, one for input_length < bucket_boundaries[0]
            and one for input_length >= bucket_boundaries[-1].
        allow_smaller_final_batch: `bool`, whether to allow a last small batch.
        dynamic_pad: `bool`, Allow variable dimensions in input shapes.
            The given dimensions are padded upon dequeue so that tensors
            within a batch have the same shapes.
        min_after_dequeue: `int`.
        num_threads: `int`. The number of threads enqueuing tensors.
        capacity: `int`, The maximum number of minibatches in the top queue,
            and also the maximum number of elements within each bucket.
        source_field: The TFRecord feature field containing the source text.
        target_field: The TFRecord feature field containing the target text.
        source_delimiter: A character to split the source text on. Defaults
          to  " " (space). For character-level training this can be set to the
          empty string.
        target_delimiter: Same as `source_delimiter` but for the target text.
    """
    IDENTIFIER = 'TFRecordSourceSequencePipeline'
    SCHEMA = TFRecordSourceSequencePipelineSchema

    def __init__(self,
                 files=None,
                 source_field='source',
                 target_field='target',
                 source_delimiter="",
                 target_delimiter="",
                 name='TFRecordSourceSequencePipeline',
                 **kwargs):
        super(TFRecordSourceSequencePipelineConfig, self).__init__(name=name, **kwargs)
        self.files = files
        self.source_field = source_field
        self.target_field = target_field
        self.source_delimiter = source_delimiter
        self.target_delimiter = target_delimiter


class ImageCaptioningPipelineSchema(BasePipelineSchema):
    files = fields.List(fields.Str(), allow_none=True)
    image_field = fields.Str(allow_none=True)
    image_format = fields.Str(allow_none=True)
    caption_ids_field = fields.Str(allow_none=True)
    caption_tokens_field = fields.Str(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ImageCaptioningPipelineConfig(**data)

    @post_dump
    def unmake(self, data):
        return ImageCaptioningPipelineConfig.remove_reduced_attrs(data)


class ImageCaptioningPipelineConfig(BasePipelineConfig):
    """An input pipeline that reads a TFRecords containing both source and target sequences.

    Args:
        name: `str`, name to give for this pipeline.
        feature_processors: `dict`, list of modules to call for each feature to be processed.
        shuffle: `bool`, If true, shuffle the data.
        num_epochs: `int`, Number of times to iterate through the dataset. If None, iterate forever.
        batch_size: The new batch size pulled from the queue (all queues will have the same size).
            If a list is passed in then each bucket will have a different batch_size.
            (python int, int32 scalar or iterable of integers of length num_buckets).
        bucket_boundaries: `list` of `int` or `None`, increasing non-negative numbers.
            The edges of the buckets to use when bucketing tensors.
            Two extra buckets are created, one for input_length < bucket_boundaries[0]
            and one for input_length >= bucket_boundaries[-1].
        allow_smaller_final_batch: `bool`, whether to allow a last small batch.
        dynamic_pad: `bool`, Allow variable dimensions in input shapes.
            The given dimensions are padded upon dequeue so that tensors
            within a batch have the same shapes.
        min_after_dequeue: `int`.
        num_threads: `int`. The number of threads enqueuing tensors.
        capacity: `int`, The maximum number of minibatches in the top queue,
            and also the maximum number of elements within each bucket.
        files: An array of file names to read from.
        image_field: The TFRecord feature field containing the source images.
        image_format: The images extensions.
        caption_ids_field: The caption ids field.
        caption_tokens_field: the caption tokends field.
    """
    IDENTIFIER = 'ImageCaptioningPipeline'
    SCHEMA = ImageCaptioningPipelineSchema

    def __init__(self,
                 files=None,
                 image_field="image/data",
                 image_format='jpg',
                 caption_ids_field="image/caption_ids",
                 caption_tokens_field="image/caption",
                 name='ImageCaptioningPipeline',
                 **kwargs):
        super(ImageCaptioningPipelineConfig, self).__init__(name=name, **kwargs)
        self.files = files
        self.image_field = image_field
        self.image_format = image_format
        self.caption_ids_field = caption_ids_field
        self.caption_tokens_field = caption_tokens_field


class PipelineSchema(BaseMultiSchema):
    __multi_schema_name__ = 'pipeline'
    __configs__ = {
        TFRecordImagePipelineConfig.IDENTIFIER: TFRecordImagePipelineConfig,
        TFRecordSequencePipelineConfig.IDENTIFIER: TFRecordSequencePipelineConfig,
        ParallelTextPipelineConfig.IDENTIFIER: ParallelTextPipelineConfig,
        TFRecordSourceSequencePipelineConfig.IDENTIFIER: TFRecordSourceSequencePipelineConfig,
        ImageCaptioningPipelineConfig.IDENTIFIER: ImageCaptioningPipelineConfig,
    }


class PipelineConfig(BaseConfig):
    SCHEMA = PipelineSchema
    IDENTIFIER = 'pipeline'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.initializations import GlorotUniformInitializerConfig, ZerosInitializerConfig
from polyaxon_schemas.layers.local import LocallyConnected1DConfig, LocallyConnected2DConfig


class TestLocalConfigs(TestCase):
    @staticmethod
    def assert_local_config(local_class, dim):
        config_dict = {
            'filters': 20,
            'kernel_size': 3,
            'strides': 1 if dim == 1 else [1, 1],
            'padding': 'valid',
            'data_format': None,
            'activation': None,
            'use_bias': True,
            'kernel_initializer': GlorotUniformInitializerConfig().to_schema(),
            'bias_initializer': ZerosInitializerConfig().to_schema(),
            'kernel_regularizer': None,
            'bias_regularizer': None,
            'activity_regularizer': None,
            'kernel_constraint': None,
            'bias_constraint': None,
        }
        if dim > 1:
            config_dict['data_format'] = None
        config = local_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_local_config(self):
        self.assert_local_config(LocallyConnected1DConfig, 1)
        self.assert_local_config(LocallyConnected2DConfig, 2)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, post_dump, post_load, validate

from polyaxon_schemas.initializations import (
    GlorotUniformInitializerConfig,
    InitializerSchema,
    ZerosInitializerConfig
)
from polyaxon_schemas.layers.base import BaseLayerConfig, BaseLayerSchema
from polyaxon_schemas.regularizations import RegularizerSchema
from polyaxon_schemas.utils import ACTIVATION_VALUES, ObjectOrListObject, StrOrFct


class LocallyConnected1DSchema(BaseLayerSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=1, max=1)
    strides = ObjectOrListObject(fields.Int, min=1, max=1, default=1, missing=1)
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    bias_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    kernel_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    bias_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    activity_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    kernel_constraint = fields.Nested(RegularizerSchema, default=None, missing=None)
    bias_constraint = fields.Nested(RegularizerSchema, default=None, missing=None)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return LocallyConnected1DConfig(**data)

    @post_dump
    def unmake(self, data):
        return LocallyConnected1DConfig.remove_reduced_attrs(data)


class LocallyConnected1DConfig(BaseLayerConfig):
    """Locally-connected layer for 1D inputs.

    The `LocallyConnected1D` layer works similarly to
    the `Conv1D` layer, except that weights are unshared,
    that is, a different set of filters is applied at each different patch
    of the input.

    Example:

    ```python
    # apply a unshared weight convolution 1d of length 3 to a sequence with
    # 10 timesteps, with 64 output filters
    x = LocallyConnected1D(64, 3)(x)
    # now x.output_shape == (None, 8, 64)
    # add a new conv1d on top
    x = LocallyConnected1D(32, 3)(x)
    # now x.output_shape == (None, 6, 32)
    ```

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number output of filters in the convolution).
        kernel_size: An integer or tuple/list of a single integer,
            specifying the length of the 1D convolution window.
        strides: An integer or tuple/list of a single integer,
            specifying the stride length of the convolution.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: Currently only supports `"valid"` (case-insensitive).
            `"same"` may be supported in the future.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to the kernel matrix.
        bias_constraint: Constraint function applied to the bias vector.

    Input shape:
        3D tensor with shape: `(batch_size, steps, input_dim)`

    Output shape:
        3D tensor with shape: `(batch_size, new_steps, filters)`
        `steps` value might have changed due to padding or strides.

    Polyaxonfile usage:

    ```yaml
    LocallyConnected1D:
      filters: 64
      kernel_size: 32
    ```
    """
    IDENTIFIER = 'LocallyConnected1D'
    SCHEMA = LocallyConnected1DSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=1,
                 padding='valid',
                 data_format=None,
                 activation=None,
                 use_bias=True,
                 kernel_initializer=GlorotUniformInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(LocallyConnected1DConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint


class LocallyConnected2DSchema(BaseLayerSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=2, max=2)
    strides = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    bias_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    kernel_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    bias_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    activity_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    kernel_constraint = fields.Nested(RegularizerSchema, default=None, missing=None)
    bias_constraint = fields.Nested(RegularizerSchema, default=None, missing=None)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return LocallyConnected2DConfig(**data)

    @post_dump
    def unmake(self, data):
        return LocallyConnected2DConfig.remove_reduced_attrs(data)


class LocallyConnected2DConfig(BaseLayerConfig):
    """Locally-connected layer for 2D inputs.

    The `LocallyConnected2D` layer works similarly
    to the `Conv2D` layer, except that weights are unshared,
    that is, a different set of filters is applied at each
    different patch of the input.

    Examples:

    ```python
    # apply a 3x3 unshared weights convolution with 64 output filters on a
    32x32 image
    # with `data_format="channels_last"`:
    x = LocallyConnected2D(64, (3, 3))(x)
    # now X.output_shape == (None, 30, 30, 64)
    # notice that this layer will consume (30*30)*(3*3*3*64) + (30*30)*64
    parameters

    # add a 3x3 unshared weights convolution on top, with 32 output filters:
    x = LocallyConnected2D(32, (3, 3))(x)
    # now x.output_shape == (None, 28, 28, 32)
    ```

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number output of filters in the convolution).
        kernel_size: An integer or tuple/list of 2 integers, specifying the
            width and height of the 2D convolution window.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        strides: An integer or tuple/list of 2 integers,
            specifying the strides of the convolution along the width and height.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        padding: Currently only support `"valid"` (case-insensitive).
            `"same"` will be supported in future.
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to the kernel matrix.
        bias_constraint: Constraint function applied to the bias vector.

    Input shape:
        4D tensor with shape:
        `(samples, channels, rows, cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(samples, rows, cols, channels)` if data_format='channels_last'.

    Output shape:
        4D tensor with shape:
        `(samples, filters, new_rows, new_cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(samples, new_rows, new_cols, filters)` if data_format='channels_last'.
        `rows` and `cols` values might have changed due to padding.

    Polyaxonfile usage:

    ```yaml
    LocallyConnected2D:
      filters: 32
      kernel_size: 3 or [3, 3]
    ```
    """
    IDENTIFIER = 'LocallyConnected2D'
    SCHEMA = LocallyConnected2DSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1),
                 padding='valid',
                 data_format=None,
                 activation=None,
                 use_bias=True,
                 kernel_initializer=GlorotUniformInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(LocallyConnected2DConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import ValidationError

from polyaxon_schemas.build import BuildConfig


class TestBuildConfigs(TestCase):
    def test_valid_image(self):
        config_dict = {
            'image': None,
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': '',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'some_image_name:sdf:sdf',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

    def test_build_config(self):
        config_dict = {
            'image': 'some_image_name',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == 'latest'
        assert config.nocache is None

    def test_build_nocache(self):
        config_dict = {
            'image': 'some_image_name',
            'nocache': True
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == 'latest'
        assert config.nocache is True

    def test_build_from_git_repo_with_install_step_config(self):
        config_dict = {
            'image': 'tensorflow:1.3.0',
            'build_steps': ['pip install tensor2tensor'],
            'env_vars': [['LC_ALL', 'en_US.UTF-8']],
            'git': 'https://github.com/tensorflow/tensor2tensor.git'
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == '1.3.0'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, post_dump, post_load, validate

from polyaxon_schemas.constraints import ConstraintSchema
from polyaxon_schemas.initializations import InitializerSchema
from polyaxon_schemas.layers.recurrent import RecurrentConfig, RecurrentSchema
from polyaxon_schemas.regularizations import RegularizerSchema
from polyaxon_schemas.utils import ACTIVATION_VALUES, ObjectOrListObject


class ConvRecurrent2DSchema(RecurrentSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=2, max=2)
    strides = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(allow_none=True,
                             validate=validate.OneOf('channels_first', 'channels_last'))
    dilation_rate = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    return_sequences = fields.Bool(default=False, missing=False)
    go_backwards = fields.Bool(default=False, missing=False)
    stateful = fields.Bool(default=False, missing=False)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ConvRecurrent2DConfig(**data)

    @post_dump
    def unmake(self, data):
        return ConvRecurrent2DConfig.remove_reduced_attrs(data)


class ConvRecurrent2DConfig(RecurrentConfig):
    """Abstract base class for convolutional recurrent layers.

    Do not use in a model -- it's not a functional layer!

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number output of filters in the convolution).
        kernel_size: An integer or tuple/list of n integers, specifying the
            dimensions of the convolution window.
        strides: An integer or tuple/list of n integers,
            specifying the strides of the convolution.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: One of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, time, ..., channels)`
            while `channels_first` corresponds to
            inputs with shape `(batch, time, channels, ...)`.
            If you never set it, then it will be "channels_last".
        dilation_rate: An integer or tuple/list of n integers, specifying
            the dilation rate to use for dilated convolution.
            Currently, specifying any `dilation_rate` value != 1 is
            incompatible with specifying any `strides` value != 1.
        return_sequences: Boolean. Whether to return the last output
            in the output sequence, or the full sequence.
        go_backwards: Boolean (default False).
            If True, rocess the input sequence backwards.
        stateful: Boolean (default False). If True, the last state
            for each sample at index i in a batch will be used as initial
            state for the sample of index i in the following batch.

    Input shape:
        5D tensor with shape `(num_samples, timesteps, channels, rows, cols)`.

    Output shape:
        - if `return_sequences`: 5D tensor with shape
            `(num_samples, timesteps, channels, rows, cols)`.
        - else, 4D tensor with shape `(num_samples, channels, rows, cols)`.

    # Masking
        This layer supports masking for input data with a variable number
        of timesteps. To introduce masks to your data,
        use an `Embedding` layer with the `mask_zero` parameter
        set to `True`.
        **Note:** for the time being, masking is only supported with Theano.

    # Note on using statefulness in RNNs
        You can set RNN layers to be 'stateful', which means that the states
        computed for the samples in one batch will be reused as initial states
        for the samples in the next batch.
        This assumes a one-to-one mapping between
        samples in different successive batches.

        To enable statefulness:
            - specify `stateful=True` in the layer constructor.
            - specify a fixed batch size for your model, by passing
                a `batch_input_size=(...)` to the first layer in your model.
                This is the expected shape of your inputs *including the batch
                size*.
                It should be a tuple of integers, e.g. `(32, 10, 100)`.

        To reset the states of your model, call `.reset_states()` on either
        a specific layer, or on your entire model.
    """
    IDENTIFIER = 'ConvRecurrent2D'
    SCHEMA = ConvRecurrent2DSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1),
                 padding='valid',
                 data_format=None,
                 dilation_rate=(1, 1),
                 return_sequences=False,
                 go_backwards=False,
                 stateful=False,
                 **kwargs):
        super(ConvRecurrent2DConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format
        self.dilation_rate = dilation_rate
        self.return_sequences = return_sequences
        self.go_backwards = go_backwards
        self.stateful = stateful


class ConvLSTM2DSchema(ConvRecurrent2DSchema):
    activation = fields.Str(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    recurrent_activation = fields.Str(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, allow_none=True)
    recurrent_initializer = fields.Nested(InitializerSchema, allow_none=True)
    bias_initializer = fields.Nested(InitializerSchema, allow_none=True)
    unit_forget_bias = fields.Bool(default=True, missing=True)
    kernel_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    recurrent_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    bias_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    kernel_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    recurrent_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    bias_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    dropout = fields.Float(default=0., missing=0., validate=validate.Range(0., 1.))
    recurrent_dropout = fields.Float(default=0., missing=0., validate=validate.Range(0., 1.))

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ConvLSTM2DConfig(**data)

    @post_dump
    def unmake(self, data):
        return ConvLSTM2DConfig.remove_reduced_attrs(data)


class ConvLSTM2DConfig(ConvRecurrent2DConfig):
    """Convolutional LSTM.

    It is similar to an LSTM layer, but the input transformations
    and recurrent transformations are both convolutional.

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number output of filters in the convolution).
        kernel_size: An integer or tuple/list of n integers, specifying the
            dimensions of the convolution window.
        strides: An integer or tuple/list of n integers,
            specifying the strides of the convolution.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: One of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, time, ..., channels)`
            while `channels_first` corresponds to
            inputs with shape `(batch, time, channels, ...)`.
            If you never set it, then it will be "channels_last".
        dilation_rate: An integer or tuple/list of n integers, specifying
            the dilation rate to use for dilated convolution.
            Currently, specifying any `dilation_rate` value != 1 is
            incompatible with specifying any `strides` value != 1.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        recurrent_activation: Activation function to use
            for the recurrent step.
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix,
            used for the linear transformation of the inputs..
        recurrent_initializer: Initializer for the `recurrent_kernel`
            weights matrix,
            used for the linear transformation of the recurrent state..
        bias_initializer: Initializer for the bias vector.
        unit_forget_bias: Boolean.
            If True, add 1 to the bias of the forget gate at initialization.
            Use in combination with `bias_initializer="zeros"`.
            This is recommended in [Jozefowicz et
              al.](http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf)
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        recurrent_regularizer: Regularizer function applied to
            the `recurrent_kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        recurrent_constraint: Constraint function applied to
            the `recurrent_kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.
        return_sequences: Boolean. Whether to return the last output
            in the output sequence, or the full sequence.
        go_backwards: Boolean (default False).
            If True, rocess the input sequence backwards.
        stateful: Boolean (default False). If True, the last state
            for each sample at index i in a batch will be used as initial
            state for the sample of index i in the following batch.
        dropout: Float between 0 and 1.
            Fraction of the units to drop for
            the linear transformation of the inputs.
        recurrent_dropout: Float between 0 and 1.
            Fraction of the units to drop for
            the linear transformation of the recurrent state.

    Input shape:
        - if data_format='channels_first'
            5D tensor with shape:
            `(samples,time, channels, rows, cols)`
        - if data_format='channels_last'
            5D tensor with shape:
            `(samples,time, rows, cols, channels)`

    Output shape:
        - if `return_sequences`
             - if data_format='channels_first'
                5D tensor with shape:
                `(samples, time, filters, output_row, output_col)`
             - if data_format='channels_last'
                5D tensor with shape:
                `(samples, time, output_row, output_col, filters)`
        - else
            - if data_format ='channels_first'
                4D tensor with shape:
                `(samples, filters, output_row, output_col)`
            - if data_format='channels_last'
                4D tensor with shape:
                `(samples, output_row, output_col, filters)`
            where o_row and o_col depend on the shape of the filter and
            the padding

    Raises:
        ValueError: in case of invalid constructor arguments.

    References:
        - [Convolutional LSTM Network: A Machine Learning Approach for
        Precipitation Nowcasting](http://arxiv.org/abs/1506.04214v1)
        The current implementation does not include the feedback loop on the
        cells output

    Polyaxonfile usage:

    ```yaml
    ConvLSTM2D:
      filters: 3
      kernel_size: 2 or [2, 2]
      strides: 1 or [1, 1]
      padding: valid
    ```
    """
    IDENTIFIER = 'ConvLSTM2D'
    SCHEMA = ConvLSTM2DSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1),
                 padding='valid',
                 data_format=None,
                 dilation_rate=(1, 1),
                 activation='tanh',
                 recurrent_activation='hard_sigmoid',
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 recurrent_initializer='orthogonal',
                 bias_initializer='zeros',
                 unit_forget_bias=True,
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 recurrent_constraint=None,
                 bias_constraint=None,
                 return_sequences=False,
                 go_backwards=False,
                 stateful=False,
                 dropout=0.,
                 recurrent_dropout=0.,
                 **kwargs):
        super(ConvLSTM2DConfig, self).__init__(
            filters,
            kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            return_sequences=return_sequences,
            go_backwards=go_backwards,
            stateful=stateful,
            **kwargs)
        self.activation = activation
        self.recurrent_activation = recurrent_activation
        self.use_bias = use_bias

        self.kernel_initializer = kernel_initializer
        self.recurrent_initializer = recurrent_initializer
        self.bias_initializer = bias_initializer
        self.unit_forget_bias = unit_forget_bias

        self.kernel_regularizer = kernel_regularizer
        self.recurrent_regularizer = recurrent_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer

        self.kernel_constraint = kernel_constraint
        self.recurrent_constraint = recurrent_constraint
        self.bias_constraint = bias_constraint

        self.dropout = dropout
        self.recurrent_dropout = recurrent_dropout



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import uuid

from unittest import TestCase

from polyaxon_schemas.job import JobConfig, JobStatusConfig, TensorboardJobConfig
from polyaxon_schemas.utils import local_now


class TestJobConfigs(TestCase):
    def test_job_config(self):
        config_dict = {
            'uuid': uuid.uuid4().hex,
            'project': 'name.name',
            'build_job': 'name.name',
            'unique_name': 'user.proj.1',
            'last_status': 'Running',
            'description': 'description',
            'config': {'k': 'v'},
            'tags': ['test'],
            'definition': None,
            'created_at': local_now().isoformat(),
            'updated_at': local_now().isoformat(),
            'started_at': local_now().isoformat(),
            'finished_at': local_now().isoformat(),
        }
        config = JobConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        config_to_dict.pop('is_clone')
        config_to_dict.pop('resources')
        config_to_dict.pop('id')
        config_to_dict.pop('total_run')
        config_to_dict.pop('user')
        config_to_dict.pop('name')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict()
        config_dict.pop('uuid')
        config_dict.pop('description')
        config_dict.pop('config')
        config_dict.pop('project')
        config_dict.pop('build_job')
        config_dict.pop('updated_at')
        config_dict.pop('definition')
        config_dict.pop('tags')
        config_to_dict.pop('id')
        config_to_dict.pop('total_run')
        config_to_dict.pop('user')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict(humanize_values=True)
        assert config_to_dict.pop('total_run') == '0s'
        assert config_to_dict.pop('created_at') == 'a few seconds ago'
        assert config_to_dict.pop('started_at') == 'a few seconds ago'
        assert config_to_dict.pop('finished_at') == 'a few seconds ago'

    def test_job_status_config(self):
        config_dict = {'id': 1,
                       'uuid': uuid.uuid4().hex,
                       'job': 1,
                       'created_at': local_now().isoformat(),
                       'status': 'Running',
                       'message': None,
                       'details': None}
        config = JobStatusConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict == config_dict

        config_dict.pop('job', None)
        config_dict.pop('uuid', None)
        config_dict.pop('details', None)
        config_to_dict = config.to_light_dict()
        assert config_to_dict == config_dict

        config_to_dict = config.to_dict(humanize_values=True)
        assert config_to_dict.pop('created_at') == 'a few seconds ago'


class TestTensorboardJobConfigs(TestCase):
    def test_job_config(self):
        config_dict = {
            'uuid': uuid.uuid4().hex,
            'project': 'name.name',
            'experiment': 1,
            'experiment_group': 1,
            'unique_name': 'user.proj.1',
            'last_status': 'Running',
            'tags': ['test'],
            'created_at': local_now().isoformat(),
            'updated_at': local_now().isoformat(),
            'started_at': local_now().isoformat(),
            'finished_at': local_now().isoformat(),
        }
        config = TensorboardJobConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        config_to_dict.pop('is_clone')
        config_to_dict.pop('resources')
        config_to_dict.pop('id')
        config_to_dict.pop('total_run')
        config_to_dict.pop('user')
        config_to_dict.pop('name')
        config_to_dict.pop('build_job')
        config_to_dict.pop('config')
        config_to_dict.pop('definition')
        config_to_dict.pop('description')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict()
        config_dict.pop('uuid')
        config_dict.pop('project')
        config_dict.pop('updated_at')
        config_dict.pop('tags')
        config_to_dict.pop('id')
        config_to_dict.pop('total_run')
        config_to_dict.pop('user')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict(humanize_values=True)
        assert config_to_dict.pop('total_run') == '0s'
        assert config_to_dict.pop('created_at') == 'a few seconds ago'
        assert config_to_dict.pop('started_at') == 'a few seconds ago'
        assert config_to_dict.pop('finished_at') == 'a few seconds ago'

    def test_job_status_config(self):
        config_dict = {'id': 1,
                       'uuid': uuid.uuid4().hex,
                       'job': 1,
                       'created_at': local_now().isoformat(),
                       'status': 'Running',
                       'message': None,
                       'details': None}
        config = JobStatusConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict == config_dict

        config_dict.pop('job', None)
        config_dict.pop('uuid', None)
        config_dict.pop('details', None)
        config_to_dict = config.to_light_dict()
        assert config_to_dict == config_dict

        config_to_dict = config.to_dict(humanize_values=True)
        assert config_to_dict.pop('created_at') == 'a few seconds ago'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from collections import Mapping

from polyaxon_schemas.build import BuildConfig
from polyaxon_schemas.exceptions import PolyaxonConfigurationError
from polyaxon_schemas.polyaxonfile.specification.base import BaseSpecification
from polyaxon_schemas.polyaxonfile.utils import cached_property


class BuildSpecification(BaseSpecification):
    """The polyaxonfile specification for build jobs.

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        ENVIRONMENT: defines the run environment for experiment.
        BUILD: defines the build step where the user can set a docker image definition
    """
    _SPEC_KIND = BaseSpecification._BUILD

    REQUIRED_SECTIONS = BaseSpecification.REQUIRED_SECTIONS + (
        BaseSpecification.BUILD,
    )
    POSSIBLE_SECTIONS = BaseSpecification.POSSIBLE_SECTIONS + (
        BaseSpecification.ENVIRONMENT, BaseSpecification.BUILD,
    )

    @cached_property
    def validated_data(self):
        return self._validated_data

    @cached_property
    def build(self):
        return self.validated_data.get(self.BUILD, None)

    @cached_property
    def environment(self):
        return self.validated_data.get(self.ENVIRONMENT, None)

    @cached_property
    def resources(self):
        return self.environment.resources if self.environment else None

    @cached_property
    def persistence(self):
        return self.environment.persistence if self.environment else None

    @cached_property
    def outputs(self):
        return self.environment.outputs if self.environment else None

    @cached_property
    def node_selectors(self):
        return self.environment.node_selectors if self.environment else None

    @classmethod
    def create_specification(cls, build_config, to_dict=True):
        if isinstance(build_config, BuildConfig):
            config = build_config.to_light_dict()
        elif isinstance(build_config, Mapping):
            config = BuildConfig.from_dict(build_config)
            config = config.to_light_dict()
        else:
            raise PolyaxonConfigurationError(
                'Create specification expects a dict or an instance of BuildConfig.')

        specification = {
            cls.VERSION: 1,
            cls.KIND: cls._SPEC_KIND,
            cls.BUILD: config
        }

        if to_dict:
            return specification
        return cls.read(specification)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from collections import Mapping


def deep_update(config, override_config):
    for k, v in six.iteritems(override_config):
        if isinstance(v, Mapping):
            k_config = config.get(k, {})
            if isinstance(k_config, Mapping):
                v_config = deep_update(k_config, v)
                config[k] = v_config
            else:
                config[k] = v
        else:
            config[k] = override_config[k]
    return config


class cached_property(object):  # noqa
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.run_exec import RunConfig


class TestRunConfigs(TestCase):
    def test_exec_config(self):
        config_dict = {
            'cmd': 'python t2t-trainer '
                   '--generate_data '
                   '--data_dir=~/t2t_data '
                   '--problems=translate_ende_wmt32k '
                   '--model=transformer '
                   '--hparams_set=transformer_base_single_gpu '
                   '--output_dir=~/t2t_train/base',
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.layers.pooling import (
    AveragePooling1DConfig,
    AveragePooling2DConfig,
    AveragePooling3DConfig,
    GlobalAveragePooling1DConfig,
    GlobalAveragePooling2DConfig,
    GlobalAveragePooling3DConfig,
    GlobalMaxPooling1DConfig,
    GlobalMaxPooling2DConfig,
    GlobalMaxPooling3DConfig,
    MaxPooling1DConfig,
    MaxPooling2DConfig,
    MaxPooling3DConfig
)


class TestPoolingConfigs(TestCase):
    @staticmethod
    def assert_pooling_config(pooling_class, dim):
        config_dict = {
            'pool_size': 1 if dim == 1 else [1, 1] if dim == 2 else [1, 1, 1],
            'strides': 1 if dim == 1 else [1, 1] if dim == 2 else [1, 1, 1],
            'padding': 'valid',
        }
        if dim > 1:
            config_dict['data_format'] = None
        config = pooling_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    @staticmethod
    def assert_global_pooling_config(pooling_class, dim):
        config_dict = {}
        if dim > 1:
            config_dict['data_format'] = None
        config = pooling_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_pooling_config(self):
        self.assert_pooling_config(MaxPooling1DConfig, 1)
        self.assert_pooling_config(AveragePooling1DConfig, 1)
        self.assert_pooling_config(MaxPooling2DConfig, 2)
        self.assert_pooling_config(AveragePooling2DConfig, 2)
        self.assert_pooling_config(MaxPooling3DConfig, 3)
        self.assert_pooling_config(AveragePooling3DConfig, 3)

    def test_global_pooling_config(self):
        self.assert_global_pooling_config(GlobalAveragePooling1DConfig, 1)
        self.assert_global_pooling_config(GlobalMaxPooling1DConfig, 1)
        self.assert_global_pooling_config(GlobalAveragePooling2DConfig, 2)
        self.assert_global_pooling_config(GlobalMaxPooling2DConfig, 2)
        self.assert_global_pooling_config(GlobalAveragePooling3DConfig, 3)
        self.assert_global_pooling_config(GlobalMaxPooling3DConfig, 3)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.authentication import AccessTokenConfig, CredentialsConfig


class TestAccessConfigs(TestCase):
    def test_access_token_config(self):
        config_dict = {'username': 'username',
                       'token': 'sdfsdf098sdf80s9dSDF800'}
        config = AccessTokenConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_credentials_config(self):
        config_dict = {'username': 'username',
                       'password': 'super-secret'}
        config = CredentialsConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig


class AccessTokenSchema(Schema):
    username = fields.Str()
    token = fields.Str()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return AccessTokenConfig(**data)

    @post_dump
    def unmake(self, data):
        return AccessTokenConfig.remove_reduced_attrs(data)


class AccessTokenConfig(BaseConfig):
    SCHEMA = AccessTokenSchema
    IDENTIFIER = 'token'

    def __init__(self, username, token):
        self.username = username
        self.token = token


class CredentialsSchema(Schema):
    username = fields.Str()
    password = fields.Str()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return CredentialsConfig(**data)

    @post_dump
    def unmake(self, data):
        return CredentialsConfig.remove_reduced_attrs(data)


class CredentialsConfig(BaseConfig):
    SCHEMA = CredentialsSchema
    IDENTIFIER = 'credentials'

    def __init__(self, username, password):
        self.username = username
        self.password = password



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from collections import Mapping

from marshmallow import Schema, ValidationError, post_dump, post_load

from polyaxon_schemas.base import BaseConfig
from polyaxon_schemas.graph import GraphConfig


class FeatureProcessorsSchema(Schema):

    @post_dump(pass_original=True)
    def make(self, data, original):
        if isinstance(original.feature_processors, Mapping):
            feature_processors = {}
            for key, val in six.iteritems(original.feature_processors):
                feature_processors[key] = val.to_dict()
            return feature_processors

        raise ValidationError("Feature processor is not valid for, `{}`".format(original))

    @post_load(pass_original=True)
    def unmake(self, data, original):
        if isinstance(original, Mapping):
            feature_processors = {}
            for key, val in six.iteritems(original):
                feature_processors[key] = GraphConfig.from_dict(val)
            return FeatureProcessorsConfig(feature_processors)

        raise ValidationError("Feature processor is not valid for, `{}`".format(original))


class FeatureProcessorsConfig(BaseConfig):
    SCHEMA = FeatureProcessorsSchema
    IDENTIFIER = 'FeatureProcessors'

    def __init__(self, feature_processors):
        self.feature_processors = feature_processors



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.exceptions import PolyaxonConfigurationError
from polyaxon_schemas.polyaxonfile.specification.base import BaseSpecification
from polyaxon_schemas.polyaxonfile.specification.build import BuildSpecification


class NotebookSpecification(BuildSpecification):
    """The polyaxonfile specification for notebooks.

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        ENVIRONMENT: defines the run environment for experiment.
        BUILD: defines the build step where the user can set a docker image definition
    """
    _SPEC_KIND = BaseSpecification._NOTEBOOK  # pylint:disable=protected-access

    def _extra_validation(self):
        try:
            super(NotebookSpecification, self)._extra_validation()
        except PolyaxonConfigurationError:
            raise PolyaxonConfigurationError(
                'NotebookSpecification must contain a valid `build` section.')


class TensorboardSpecification(BuildSpecification):
    """The polyaxonfile specification for tensorboard.

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        ENVIRONMENT: defines the run environment for experiment.
        BUILD: defines the build step where the user can set a docker image definition
    """
    _SPEC_KIND = BaseSpecification._TENSORBOARD  # pylint:disable=protected-access

    def _extra_validation(self):
        try:
            super(TensorboardSpecification, self)._extra_validation()
        except PolyaxonConfigurationError:
            raise PolyaxonConfigurationError(
                'TensorboardSpecification must contain a valid `build` section.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function


def get_task_configs(cluster, is_distributed, configs, default_config, task_type):
    result_configs = {}
    if not is_distributed:
        return result_configs

    for session_config in configs or []:
        result_configs[session_config.index] = session_config

    if default_config:
        for i in range(cluster.get(task_type, 0)):
            result_configs[i] = result_configs.get(i, default_config)

    return result_configs


def get_task_job_resources(cluster, is_distributed, resources, default_resources, task_type):
    if not is_distributed:
        return None

    result_resources = {}
    for resources_config in resources or []:
        result_resources[resources_config.index] = resources_config

    if default_resources:
        for i in range(cluster.get(task_type, 0)):
            result_resources[i] = result_resources.get(i, default_resources)

    return result_resources


def get_task_job_node_selectors(cluster,
                                is_distributed,
                                node_selectors,
                                default_node_selectors,
                                task_type):
    if not is_distributed:
        return None

    result_node_selectors = {}
    for node_selectors_config in node_selectors or []:
        result_node_selectors[node_selectors_config.pop('index', 0)] = node_selectors_config

    if default_node_selectors:
        for i in range(cluster.get(task_type, 0)):
            result_node_selectors[i] = result_node_selectors.get(i, default_node_selectors)

    return result_node_selectors



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema
from polyaxon_schemas.bridges import BridgeSchema
from polyaxon_schemas.graph import GraphSchema
from polyaxon_schemas.losses import LossSchema
from polyaxon_schemas.metrics import MetricSchema
from polyaxon_schemas.optimizers import OptimizerSchema
from polyaxon_schemas.utils import ObjectOrListObject


class BaseModelSchema(Schema):
    graph = fields.Nested(GraphSchema)
    loss = fields.Nested(LossSchema, allow_none=True)
    optimizer = fields.Nested(OptimizerSchema, allow_none=True)
    metrics = fields.Nested(MetricSchema, many=True, allow_none=True)
    summaries = ObjectOrListObject(fields.Str, allow_none=True)
    clip_gradients = fields.Float(allow_none=True)
    clip_embed_gradients = fields.Float(allow_none=True)
    name = fields.Str(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return BaseModelConfig(**data)

    @post_dump
    def unmake(self, data):
        return BaseModelConfig.remove_reduced_attrs(data)


class BaseModelConfig(BaseConfig):
    SCHEMA = BaseModelSchema
    IDENTIFIER = 'Model'
    REDUCED_ATTRIBUTES = ['name', 'summaries']

    def __init__(self,
                 graph,
                 loss=None,
                 optimizer=None,
                 metrics=None,
                 summaries=None,
                 clip_gradients=0.5,
                 clip_embed_gradients=0.,
                 name=None):
        self.graph = graph
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics
        self.summaries = summaries
        self.clip_gradients = clip_gradients
        self.clip_embed_gradients = clip_embed_gradients
        self.name = name


class ClassifierSchema(BaseModelSchema):
    one_hot_encode = fields.Bool(allow_none=True)
    n_classes = fields.Int(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ClassifierConfig(**data)

    @post_dump
    def unmake(self, data):
        return ClassifierConfig.remove_reduced_attrs(data)


class ClassifierConfig(BaseModelConfig):
    """Classifier base model.

    Args(programmatic):
        mode: `str`, Specifies if this training, evaluation or prediction. See `Modes`.
        graph_fn: Graph function. Follows the signature:
            * Args:
                * `mode`: Specifies if this training, evaluation or prediction. See `Modes`.
                * `inputs`: the feature inputs.

    Args(polyaxonfile):
        graph: Graph definition. see [Graph]()

    Args(shared):
        loss: An instance of `LossConfig`. Default value `SigmoidCrossEntropyConfig`.
        optimizer: An instance of `OptimizerConfig`.
            Default value `AdamConfig(learning_rate=0.001)`.
        metrics: a list of `MetricConfig` instances.
        summaries: `str` or `list`. The verbosity of the tensorboard visualization.
            Possible values:
             [`all`, `activations`, `loss`, `learning_rate`, `variables`, `gradients`]
        clip_gradients: `float`. Gradients  clipping by global norm.
        clip_embed_gradients: `float`. Embedding gradients clipping to a specified value.
        one_hot_encode: `bool`. to one hot encode the outputs.
        n_classes: `int`. The number of classes used in the one hot encoding.
        name: `str`, the name of this model, everything will be encapsulated inside this scope.

    Returns:
        `EstimatorSpec`

    Programmatic usage:

    ```python
    def graph_fn(mode, features):
        x = plx.layers.Conv2D(filters=32, kernel_size=5)(features['image'])
        x = plx.layers.MaxPooling2D(pool_size=2)(x)
        x = plx.layers.Conv2D(filters=64, kernel_size=5)(x)
        x = plx.layers.MaxPooling2D(pool_size=2)(x)
        x = plx.layers.Flatten()(x)
        x = plx.layers.Dense(units=10)(x)
        return x

    model = plx.models.Classifier(
        mode=mode,
        graph_fn=graph_fn,
        loss=SigmoidCrossEntropyConfig(),
        optimizer=AdamConfig(
            learning_rate=0.007, decay_type='exponential_decay', decay_rate=0.1),
        metrics=[AccuracyConfig(), PrecisionConfig()],
        summaries='all',
        one_hot_encode=True,
        n_classes=10)
    ```

    Polyaxonfile usage:

    ```yaml
    model:
      classifier:
        loss: SigmoidCrossEntropy
        optimizer:
          Adam:
            learning_rate: 0.007
            decay_type: exponential_decay
            decay_rate: 0.2
        metrics:
          - Accuracy
          - Precision
        one_hot_encode: true
        n_classes: 10
        graph:
          input_layers: image
          layers:
            - Conv2D:
                filters: 32
                kernel_size: 5
                strides: 1
            - MaxPooling2D:
                pool_size: 2
            - Conv2D:
                filters: 64
                kernel_size: 5
            - MaxPooling2D:
                pool_size: 2
            - Flatten:
            - Dense:
                units: 1024
                activation: tanh
            - Dense:
               units: 10
    ```

    or use model_type to reduce the nesting level

    ```yaml
    model:
      model_type: classifier
      loss: SigmoidCrossEntropy
      optimizer:
        Adam:
          learning_rate: 0.007
          decay_type: exponential_decay
          decay_rate: 0.2
      metrics:
        - Accuracy
        - Precision
      one_hot_encode: true
      n_classes: 10
      graph:
        input_layers: image
        layers:
          - Conv2D:
              filters: 32
              kernel_size: 5
              strides: 1
          - MaxPooling2D:
              pool_size: 2
          - Conv2D:
              filters: 64
              kernel_size: 5
          - MaxPooling2D:
              pool_size: 2
          - Flatten:
          - Dense:
              units: 1024
              activation: tanh
          - Dense:
             units: 10
    ```
    """
    SCHEMA = ClassifierSchema
    IDENTIFIER = 'Classifier'

    def __init__(self,
                 graph,
                 loss=None,
                 optimizer=None,
                 metrics=None,
                 summaries=None,
                 clip_gradients=0.5,
                 clip_embed_gradients=0.,
                 one_hot_encode=None,
                 n_classes=None,
                 name=None,):
        super(ClassifierConfig, self).__init__(
            graph=graph,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            summaries=summaries,
            clip_gradients=clip_gradients,
            clip_embed_gradients=clip_embed_gradients,
            name=name)
        self.one_hot_encode = one_hot_encode
        self.n_classes = n_classes


class RegressorSchema(BaseModelSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return RegressorConfig(**data)

    @post_dump
    def unmake(self, data):
        return RegressorConfig.remove_reduced_attrs(data)


class RegressorConfig(BaseModelConfig):
    """Regressor base model.

    Args(programmatic):
        mode: `str`, Specifies if this training, evaluation or prediction. See `Modes`.
        graph_fn: Graph function. Follows the signature:
            * Args:
                * `mode`: Specifies if this training, evaluation or prediction. See `Modes`.
                * `inputs`: the feature inputs.

    Args(polyaxonfile):
        graph: Graph definition. see [Graph]()

    Args(both):
        loss: An instance of `LossConfig`. Default value `MeanSquaredErrorConfig`.
        optimizer: An instance of `OptimizerConfig`.
            Default value `AdamConfig(learning_rate=0.001)`.
        metrics: a list of `MetricConfig` instances.
        summaries: `str` or `list`. The verbosity of the tensorboard visualization.
            Possible values:
             [`all`, `activations`, `loss`, `learning_rate`, `variables`, `gradients`]
        clip_gradients: `float`. Gradients  clipping by global norm.
        clip_embed_gradients: `float`. Embedding gradients clipping to a specified value.
        name: `str`, the name of this model, everything will be encapsulated inside this scope.

    Returns:
        `EstimatorSpec`

    Programmatic usage:

    ```python
    def graph_fn(mode, features):
        x = features['x']
        x = plx.layers.LSTM(units=10)(x)
        return plx.layers.Dense(units=1)(x)

    model = plx.models.Regressor(
        mode=mode,
        graph_fn=graph_fn,
        loss=MeanSquaredErrorConfig(),
        optimizer=AdagradConfig(learning_rate=0.1),
        metrics=[
            RootMeanSquaredErrorConfig(),
            MeanAbsoluteErrorConfig()])
    ```

    Polyaxonfile usage:

    ```yaml
    model:
      regressor:
        loss: MeanSquaredError
        optimizer:
          Adagrad:
            learning_rate: 0.1
        metrics:
          - RootMeanSquaredError
          - MeanAbsoluteError
        graph:
          input_layers: x
          layers:
            - LSTM:
                units: 19
            - Dense:
                units: 1
    ```

    or use model_type to reduce the nesting level

    ```yaml
    model:
      model_type: regressor:
      loss: MeanSquaredError
      optimizer:
        Adagrad:
          learning_rate: 0.1
      metrics:
        - RootMeanSquaredError
        - MeanAbsoluteError
      graph:
        input_layers: x
        layers:
          - LSTM:
              units: 19
          - Dense:
              units: 1
    ```
    """
    SCHEMA = RegressorSchema
    IDENTIFIER = 'Regressor'


class GeneratorSchema(BaseModelSchema):
    encoder = fields.Nested(GraphSchema)
    decoder = fields.Nested(GraphSchema)
    bridge = fields.Nested(BridgeSchema)

    class Meta:
        ordered = True
        exclude = ('graph',)

    @post_load
    def make(self, data):
        return GeneratorConfig(**data)

    @post_dump
    def unmake(self, data):
        return GeneratorConfig.remove_reduced_attrs(data)


class GeneratorConfig(BaseModelConfig):
    """Generator base model.

    Args(programmatic):
        mode: `str`, Specifies if this training, evaluation or prediction. See `Modes`.
        encoder_fn: Encoder Graph function. Follows the signature:
            * Args:
                * `mode`: Specifies if this training, evaluation or prediction. See `Modes`.
                * `inputs`: the feature inputs.
        decoder_fn: Decoder Graph function. Follows the signature:
            * Args:
                * `mode`: Specifies if this training, evaluation or prediction. See `Modes`.
                * `inputs`: the feature inputs.
        bridge_fn: The bridge to use. Follows the signature:
            * Args:
                * `mode`: Specifies if this training, evaluation or prediction. See `Modes`.
                * `inputs`: the feature inputs.
                * `encoder_fn`: the encoder function.
                * `decoder_fn` the decoder function.
    Args(polyaxonfile):
        encoder: Graph definition. see [Graph]()
        decoder: Graph definition. see [Graph]()
        bridge: Graph definition. see [Graph]()

    Args(shared):
        loss: An instance of `LossConfig`. Default value `SigmoidCrossEntropyConfig`.
        optimizer: An instance of `OptimizerConfig`.
            Default value `AdadeltaConfig(learning_rate=0.4)`.
        summaries: `str` or `list`. The verbosity of the tensorboard visualization.
            Possible values: `all`, `activations`, `loss`, `learning_rate`, `variables`, `gradients`
        metrics: a list of `MetricConfig` instances.
        summaries: `str` or `list`. The verbosity of the tensorboard visualization.
            Possible values: `all`, `activations`, `loss`, `learning_rate`, `variables`, `gradients`
        clip_gradients: `float`. Gradients  clipping by global norm.
        clip_embed_gradients: `float`. Embedding gradients clipping to a specified value.
        name: `str`, the name of this model, everything will be encapsulated inside this scope.

    Returns:
        `EstimatorSpec`

    Programmatic usage:

    ```python
    def encoder_fn(mode, features):
    x = plx.layers.Dense(units=128)(features)
    x = plx.layers.Dense(units=256)(x)
    return x


    def decoder_fn(mode, features):
        x = plx.layers.Dense(units=256)(features)
        return plx.layers.Dense(units=784)(x)


    def bridge_fn(mode, features, labels, loss, encoder_fn, decoder_fn):
        return plx.bridges.NoOpBridge(mode)(features, labels, loss, encoder_fn, decoder_fn)

    model = plx.models.Generator(
        mode=mode,
        encoder_fn=encoder_fn,
        decoder_fn=decoder_fn,
        bridge_fn=bridge_fn,
        loss=MeanSquaredErrorConfig(),
        optimizer=AdadeltaConfig(learning_rate=0.9),
        summaries=['loss'])
    ```

    Polyaxonfile usage:

    ```yaml
    model:
      generator:
        loss:
          MeanSquaredError:
        optimizer:
          Adam:
            learning_rate: 0.9
        metrics:
          - Accuracy
        bridge: NoOpBridge
        encoder:
          input_layers: image
          layers:
            - Dense:
                units: 128
            - Dense:
                units: 256
                name: encoded
        decoder:
          input_layers: encoded
          layers:
            - Dense:
                units: 256
            - Dense:
                units: 784
    ```

    or use model_type to reduce the nesting level

    ```yaml
    model:
      model_type: generator:
      loss:
        MeanSquaredError:
      optimizer:
        Adam:
          learning_rate: 0.9
      metrics:
        - Accuracy
      bridge: NoOpBridge
      encoder:
        input_layers: image
        layers:
          - Dense:
              units: 128
          - Dense:
              units: 256
              name: encoded
      decoder:
        input_layers: encoded
        layers:
          - Dense:
              units: 256
          - Dense:
              units: 784
    ```
    """
    SCHEMA = GeneratorSchema
    IDENTIFIER = 'Generator'

    def __init__(self,
                 encoder,
                 decoder,
                 bridge,
                 loss=None,
                 optimizer=None,
                 metrics=None,
                 summaries=None,
                 clip_gradients=0.5,
                 clip_embed_gradients=0.,
                 name=None,):
        super(GeneratorConfig, self).__init__(
            graph=None,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            summaries=summaries,
            clip_gradients=clip_gradients,
            clip_embed_gradients=clip_embed_gradients,
            name=name)

        self.encoder = encoder
        self.decoder = decoder
        self.bridge = bridge


class ModelSchema(BaseMultiSchema):
    __multi_schema_name__ = 'model'
    __configs__ = {
        BaseModelConfig.IDENTIFIER: BaseModelConfig,
        ClassifierConfig.IDENTIFIER: ClassifierConfig,
        RegressorConfig.IDENTIFIER: RegressorConfig,
        GeneratorConfig.IDENTIFIER: GeneratorConfig,
    }
    __support_snake_case__ = True


class ModelConfig(BaseConfig):
    SCHEMA = ModelSchema
    IDENTIFIER = 'model'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.version import (
    ChartVersionConfig,
    CliVersionConfig,
    LibVersionConfig,
    PlatformVersionConfig
)


class TestVersionConfigs(TestCase):
    def test_cli_version_config(self):
        config_dict = {'latest_version': '2.2.2', 'min_version': '1.1.0'}
        config = CliVersionConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_platform_version_config(self):
        config_dict = {'latest_version': '2.2.2', 'min_version': '1.1.0'}
        config = PlatformVersionConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_lib_version_config(self):
        config_dict = {'latest_version': '2.2.2', 'min_version': '1.1.0'}
        config = LibVersionConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_chart_version_config(self):
        config_dict = {'version': '2.2.2'}
        config = ChartVersionConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import random
import uuid

from unittest import TestCase

from polyaxon_schemas.clusters import ClusterNodeConfig, NodeGPUConfig, PolyaxonClusterConfig


class TestClusterConfigs(TestCase):

    @staticmethod
    def create_gpu():
        return {
            'uuid': uuid.uuid4().hex,
            'serial': 'serial',
            'name': 'gamma',
            'index': random.randint(1, 100),
            'memory': 10,
            'cluster_node': uuid.uuid4().hex
        }

    @classmethod
    def create_cluster_node(cls):
        return {
            'uuid': uuid.uuid4().hex,
            'sequence': 1,
            'name': 'node',
            'hostname': 'hostname',
            'role': 'master',
            'docker_version': 'v1',
            'kubelet_version': 'v1',
            'os_image': 'some image',
            'kernel_version': 'v1',
            'schedulable_taints': True,
            'schedulable_state': True,
            'memory': 10,
            'cpu': 2,
            'n_gpus': 3,
            'status': 'Running',
            'gpus': [cls.create_gpu()]
        }

    def test_node_gpu_config(self):
        config_dict = self.create_gpu()
        config = NodeGPUConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_cluster_node_config(self):
        config_dict = self.create_cluster_node()
        config = ClusterNodeConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_plx_cluster_config(self):
        config_dict = {
            'version_api': {},
            'nodes': [self.create_cluster_node(), self.create_cluster_node()]
        }
        config = PolyaxonClusterConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig
from polyaxon_schemas.hooks import HookSchema
from polyaxon_schemas.processing.pipelines import PipelineSchema


class TrainSchema(Schema):
    data_pipeline = fields.Nested(PipelineSchema)
    steps = fields.Int(allow_none=True)
    hooks = fields.Nested(HookSchema, many=True, allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return TrainConfig(**data)

    @post_dump
    def unmake(self, data):
        return TrainConfig.remove_reduced_attrs(data)


class TrainConfig(BaseConfig):
    SCHEMA = TrainSchema
    IDENTIFIER = 'train'

    def __init__(self, data_pipeline, steps=100, hooks=None):
        self.data_pipeline = data_pipeline
        self.hooks = hooks
        self.steps = steps



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig


class RunSchema(Schema):
    cmd = fields.Str()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return RunConfig(**data)

    @post_dump
    def unmake(self, data):
        return RunConfig.remove_reduced_attrs(data)


class RunConfig(BaseConfig):
    SCHEMA = RunSchema
    IDENTIFIER = 'run'

    def __init__(self, cmd):
        self.cmd = cmd
